---
--    Copyright (c) 2015 by SAS Institute Inc., Cary, NC, USA.

---    This module contains general utility functions related to SAS data.
--    This module is meant to fit on top of the SAS API provided by the sasext module
--    to make a more user-friendly interface to general data-portions of SAS.
-- @module sasdata
local filesys = require 'sas.risk.utils.filesys'
local stringutils = require 'sas.risk.utils.stringutils'
local args = require 'sas.risk.utils.args'
local errors = require 'sas.risk.utils.errors'
local sas_msg = require 'sas.risk.utils.sas_msg'

local M={}


--- Check for the existence of a SAS libref
-- @param libref Name of a SAS libref
-- @return true or false whether the given libref exists (is allocated in SAS)
function M.libref_exists (libref)
   return sas.libref(libref) == 0   -- rc of 0 from SAS means the libref is assigned
end

--- Require the existence of a SAS libref. Prints error message and quits if no libref exists.
-- @param libref Name of a SAS libref
function M.require_libref(libref)
   if not M.libref_exists(libref) then
      errors.throw('utils_sasdata_required_library_missing', tostring(libref))
   end
end

--- Gets the file path associated with a libref
-- @param libref  The SAS libref to extract the path from
-- @return the last path assigned to the libref with quotes removed
function M.get_path_from_libref(libref)
   local path = sas.pathname(libref, 'L')
   path = sas.tranwrd(path, '(', '')
   path = sas.tranwrd(path, ')', '')
   path = sas.dequote(sas.left(sas.trim(path)))
   return path
end

local free_one = function(libref)
   if M.libref_exists(libref) then
      sas.glibname(libref)  -- we should always use sas.glibname() for our libname needs
      sas_msg.print('utils_sasdata_library_deallocated', string.upper(libref))
   end
end

local nextTempNumber=1

---
--   Allocates a SAS library and assigns the given libref to it.
--   The directory specified can be either a directory or a parenthesis-surrounded,
--   space delimited list of librefs to create a concatenated library.
-- 
--   This function throws an error when the library is not allocated successfully
--   
-- @param libref  The SAS libref to allocate to the given directory or nil if any generated libref name will do
-- @param[opt] dir The directory to allocate the libref to or nil if the libref is to be deallocated
-- @param[opt] libref_options A libref_options string for the SAS libref/library being defined as documented in the SAS documentation
-- @param[opt] options {} of options:  ensure_dir (if true ensures the directory)
-- @return the libref when allocating the library, nil when deallocating the library
function M.libname(libref, dir, libref_options, options)
   -- Dealloc the library first if it exists, so that we can definitively determine
   -- when the libref failed without checking the assorted non-zero RCs considered "OK".
   -- This also ensures that the old library assignment won't exist after this ftn returns.
   if libref then
      if type(libref) == "string" then
         free_one(libref)
      elseif type(libref) == "table" then
         for i,v in ipairs(libref) do
            free_one(v)
         end
         return -- we currently don't define assigning a libref array, only freeing a libref array
      end
      if not dir then
         return
      end
   else
      repeat
         libref="_lib" .. nextTempNumber
         nextTempNumber=nextTempNumber+1
      until not M.libref_exists(libref)
   end
   
   if options and options.ensure_dir then
      filesys.ensure_dir_exists(dir)
   end
   
   sas.glibname(libref, dir, libref_options)  -- we should always use sas.glibname() for our libname needs

   if not M.libref_exists(libref) then
      errors.throw("utils_sasdata_libref_alloc_failure", libref, dir, libref_options)
   end

   if not sas.get_option("GENERIC") then 
     if libref_options and #libref_options > 0 then
        sas_msg.print('utils_sasdata_libref_allocd_with_opts', libref, dir, libref_options)
     else
        sas_msg.print('utils_sasdata_libref_allocd', libref, dir)
     end
   end
   
   return libref
end

--- Get a list of data sets that are in a libref. Works with concatenated librefs.
-- @param libref a string containing the name of the libref
-- @return A list of data sets in the form 'dataset1 dataset2 dataset3'
function M.dslist(libref)
   if M.libref_exists(libref) then
      return table.concat(sas.memlist(libref,"DATA", true),' ')
   else
      return nil
   end
end

--- Wrapper around sas.memlist to return an array of strings of the names of the members of the given library.
--  This function also allows a directory to be given instead of a libref (which will cause a libref to be temporarily allocated)
-- @param libref  SAS libref or directory 
-- @param[opt] type  DATA, VIEW or ALL values are valid
-- @return Members of the libref {'dataset1', 'dataset2', 'dataset3'}
-- @usage 
-- local lst = sasdata.member_list('saswork')  -- produce a list of everything in WORK
-- local lst_ds = sasdata.member_list('sashelp','DATA')  -- produce a list of data sets in SASHELP
--   -- produce a list of data sets in the directory c:/temp
-- local lst_dir = sasdata.member_list('c:/temp/','DATA') 
function M.member_list(libref, type)
  type=type or 'ALL'
  local members
  if (string.len(libref) <= 8 and M.libref_exists(libref)) then
    members = sas.memlist(libref, type, true)
  else  -- specified a directory instead of a libref 
    
    if not (sas.fileexist(libref) == 1) then return nil end
    local generated_libref = M.libname(nil, libref, nil)
    members = sas.memlist(generated_libref, type, true)
    libname(generated_libref) -- free it
  end
  return members
end

-- alias sasdata.open to sasext.open
M.open = sas.open

--- Check for the existence of a data set
-- @param dsname Fully qualified name of the data set
-- @return boolean true if the data set exists
function M.ds_exists(dsname)
   return sas.exists(dsname, 'DATA')
end

--- Check for the existence of a data set in directory
-- @param dir 
-- @param datasetName Name of the data set
-- @return boolean true if the data set exists
function M.ds_exists_in_dir(dir, datasetName)
   args.assert_string(dir)
   args.assert_string(datasetName)

   if not filesys.dir_exists(dir) then
      return false
   end
   local libref = M.libname(nil, dir)
   local rc = M.ds_exists(libref .. '.' .. datasetName)
   free_one(libref)
   return rc
end

--- Require the existence of a data set
-- @param dsname Fully qualified name of the data set
-- @return returns if the data set exists or terminates with a message
function M.require_ds(dsname)
   if M.ds_exists(dsname) == false then
      errors.throw('utils_sasdata_required_dataset_missing', tostring(dsname))
   end
end

--- Convenience method to delete a data set and handle nil/empty and non-existent situations
-- @param dsname Fully qualified name of the data set
function M.delete_ds(dsname)
   if dsname and #dsname > 0 and M.ds_exists(dsname) then
      sas.delete(dsname)
   end
end



--- Check for the existence of a variable in a data set
-- @param dsname Fully qualified name of the data set
-- @param var_name Name of the variable to look for
-- @return returns boolean indicating whether the variable exists or not
function M.var_exists(dsname, var_name)

    local rc=false
    --print(ds)
    local dsid = sas.open(dsname,"i")

    if dsid == 0 then
       print(sas.sysmsg());  -- print this in case there's something useful in it from the call to open()
    else
       local varnum = sas.varnum(dsid, var_name)
       --print("varnum for " .. tostring(var_name) .. " is " .. tostring(varnum))
       rc = varnum ~= 0  -- the variable exists if told the varnum is not zero
       sas.close(dsid)
    end
    --print("var_exists() is returning " .. tostring(rc))
    return rc
end


--- Enforces that the given variable exists in the given data set.
-- @param dsname Fully qualified name of the data set
-- @param var_name Name of the variable to look for
-- @return Nothing if the variable exists. If it does not exist, an error is thrown.
function M.require_var(dsname, var_name)
   if not M.var_exists(dsname, var_name) then
      errors.throw("utils_sasdata_required_var_does_not_exist", dsname, var_name)
   end
end


--- Add new variables to data set. @param cols table containing tables, one for each variable, as follows:
-- @param dsname Fully qualified name of the data set
-- @param cols table containing tables, one for each variable, as follows:
-- { var1={type='C', length=25, label="this is var1", format= ...},
--   var2={type='N', length=8,   label="this is var2", format= ...}}
function M.add_vars(ds, cols)
   sas.submit_[[
         proc sql noprint;
   ]]
   for name,col in pairs(cols) do
      sas.submit_[[alter table @ds@]]
      local type = 'NUM'
      if col.type=='C' then
         type = 'CHAR'
      end
      local length = col.length or 8
      local label= col.label or ""

      sas.submit_[[ add @name@ @type@ length=@length@ label="@label@" ]]
      if col.format then
         sas.submit_([[ format = @format@]], {format=col.format})
      end
      if col.informat then
         sas.submit_([[ informat=@informat@]], {informat=col.informat})
      end
      sas.submit_(";")
   end
   sas.submit("quit;")
end


--- Deletes the given list of data set names
-- @param libref 
-- @param dsnames An array of names as either a space separated string or an table {'ds1','ds2', ...}
function M.delete_datasets(libref, ds_names)
   args.assert_string(libref)
   if ds_names == nil then
      return
   end
   local datasets
   if type(ds_names) == "string" then
      datasets = stringutils.split(ds_names," ")
   elseif type(ds_names) == "table" then
      datasets = ds_names
   end
   assert(datasets, "invalid list of data set names")
   for i,v in ipairs(datasets) do
      sas.delete(libref .. "." .. v)
   end
end

--- Copy a schema for a specific data set from one library to another. New data set has no observations
-- @param inlib The library to copy the schema from
-- @param outlib The place to put the copied schema
-- @param[opt] noconstraints is optional - true -> turn off constraints
--- @usage
--  create a empty data set in work with the same fields
--     local inlib = string.gsub(table, '%..*','') 
--     local memname =  string.gsub(table, '.*%.','')    
 --    sasdata.copy_ds_schema(inlib, 'work', memname, true)
function M.copy_ds_schema(inlib, outlib, ds_name, noconstraints)
      ---
      -- Save the current OBS system option and then set it to zero so    
      -- that no data is copied                                           
      --                                                                 
      -- Use PROC COPY to create an empty data set with exactly the same  
      -- attributes, indexes (if possible), and constraints (if possible), then restore the OBS       
      -- system option 
      -- 
      local constraints =  "constraint=yes" 
      if noconstraints then
         constraints =  "constraint=no" 
      end                                                 

      sas.submit[[
      %let SAVED_OBS=%sysfunc(getoption(obs, KEYWORD));
      options obs=0;
      proc copy in=@inlib@ out=@outlib@ @constraints@ index=yes;
         select @ds_name@;
      run;
      options &SAVED_OBS;
      ]]
end



---
-- Copies a single data set from one library to another, optionally filtering it with a given SQL WHERE clause,
-- and preserving its indexes. Constraints are preserved as best possible per the constraint=yes option
-- documented at  http://support.sas.com/documentation/cdl/en/proc/63079/HTML/default/viewer.htm#p1juxu16zautpxn1dikxecc3kn7w.htm
-- @param inlib Input libname
-- @param outlib Output libname
-- @param dsname Data set name (source and destination)
-- @param filter SQL WHERE clause
-- @param[opt] drop_list table of variables to drop 
function M.copy_and_filter(inlib, outlib, ds_name, filter, drop_list)
   if filter == nil then
      sas.submit([[
         proc copy in=@inlib@ out=@outlib@ constraint=yes index=yes;
            select @ds_name@;
         run;
      ]], {inlib=inlib, outlib=outlib, ds_name=ds_name})
   else
   
      M.copy_ds_schema(inlib, outlib, ds_name)
   
      sas.submit([[
      proc sql noprint;
        insert into @outlib@.@ds_name@
           select *
           from @inlib@.@ds_name@
           @filter@
           ;
      quit;
      ]], {inlib=inlib, outlib=outlib, ds_name=ds_name, filter=filter})

   end
   
   if drop_list then
      local drop_cols = table.concat(drop_list, ' ')
      sas.submit[[
         data @outlib@.@ds_name@;
            set @outlib@.@ds_name@;
            drop @drop_cols@;
         run;]]
   end      
end


---
-- Copies an entire library of data sets.
-- @param from_lib Input libname
-- @param to_lib Output libname
-- @param options   a string containing any of these options: <br/>
-- http://support.sas.com/documentation/cdl/en/proc/63079/HTML/default/viewer.htm#p1juxu16zautpxn1dikxecc3kn7w.htm
-- @param[opt] selected_data_sets  Table of data set names
-- @param[opt] excluded_data_sets  Table of excluded data sets
function M.copy_lib(from_lib, to_lib, options, selected_data_sets, excluded_data_sets)
   options = options or 'CONSTRAINT=YES'
   args.assert_table_or_nil(selected_data_sets,4)
   args.assert_table_or_nil(excluded_data_sets,5)

   sas.submit_ [[
         proc copy in=@from_lib@ out=@to_lib@ @options@;
   ]]
   if selected_data_sets then
      local ds_list_str = table.concat(selected_data_sets, " ")
      sas.submit_[[
            select @ds_list_str@;
      ]]
   end
   if excluded_data_sets then
      local ds_list_str = table.concat(excluded_data_sets, " ")
      sas.submit_[[
            exclude @ds_list_str@;
      ]]
   end
   sas.submit("run;")
end

--- Sort Data set - Uses Proc Sort
-- @param in_ds Fully qualified data set name for the input
-- @param[opt] out_ds Fully qualified data set name for the input. If nil the input data set is used.
-- @param by_cols Table of sort variables or a space delimited string
-- @param[opt] keywords Options to the Sort procedure
function M.sort_ds(in_ds, out_ds, by_cols, keywords)

   keywords = keywords or ""

   -- if cols are not in a string (can be space delimited values), must be in a list (table)
   if type(by_cols) == "table" then
      by_cols = table.concat(by_cols, " ")
   end

   if out_ds==nil then
      sas.submit([[proc sort data=@in_ds@              @keywords@;
                      by @by_cols@;
                   run;]],
                 {in_ds=in_ds,                by_cols=by_cols, keywords=keywords})
   else
      sas.submit([[proc sort data=@in_ds@ out=@out_ds@ @keywords@;
                      by @by_cols@;
                   run;]],
                 {in_ds=in_ds, out_ds=out_ds, by_cols=by_cols, keywords=keywords})
   end
end


--- Perform an SQL statement via proc sql.
-- @param sql_stmt SQL statement string with substition fields
-- @param substitution_args Table of name value pairs to substitute in the sql_stmt string
-- @param[opt] proc_sql_options Noprint is used by default
-- @param[opt] addl_lookup_level Additional stack levels to resolve variables for substitution. 0 is the default
--@return If PROC SQL diagnoses an error, a Lua error is thrown
--@usage
-- sasdata.sql("create table " .. dsname .. " as\n" .. selectStmt .. ";",   
--        nil, nil, 1) -- add this function as a single level to what caller passed
function M.sql (sql_stmt, substitution_args, proc_sql_options, addl_lookup_level)
   addl_lookup_level=addl_lookup_level or 0
   proc_sql_options = proc_sql_options or "noprint"  -- default to noprint if options not specified

   sas.submit("proc sql " .. proc_sql_options .. ";\n" ..
                   sql_stmt .. ";\n" ..  -- add a benign semicolon in case the caller didn't include one
                "quit;", substitution_args, addl_lookup_level+1) -- add this function as a single level to what caller passed

   -- Now check the SQLRC and see if the proc reported an error, if so throw a Lua error.
   -- SQL automatic variables are described here:
   --     http://support.sas.com/documentation/cdl/en/sqlproc/63043/HTML/default/viewer.htm#p0xlnvl46zgqffn17piej7tewe7p.htm
   local sqlrc = sas.symget("SQLRC") -- returns a number!
   if sqlrc > 4 then
      errors.throw("utils_sasdata_error_running_proc_sql", tostring(sqlrc))
   end
end


--- Queries a whole table and returns a lua table 
-- Should only be used on tables we expect we can keep in memory
-- @param selectStmt SQL select statement string with substition fields
-- @param substitution_args Table of name value pairs to substitute in the sql_stmt string
-- @param[opt] trim_string_values Boolean to trim character values in the returned lua table
-- @param[opt] addl_lookup_level Additional stack levels to resolve variables for substitution. 0 is the default
--@return Table of tables (outer table is rows(1..n), inner table is named columns for each for row). If PROC SQL diagnoses an error, a Lua error is thrown
function M.select(selectStmt, substitution_args, trim_string_values, addl_lookup_level)
   addl_lookup_level=addl_lookup_level or 0
   trim_string_values = trim_string_values or true  -- default to true if not specified

   local dsname = "WORK._select"
   M.delete_ds(dsname)

   M.sql("create table " .. dsname .. " as\n" ..
        selectStmt .. ";",   -- add a benign semicolon in case the caller didn't include one
        substitution_args, nil, addl_lookup_level+1) -- add this function as a single level to what caller passed

   local dsid = M.open(dsname)
   local col_names = {}
   local col_count = sas.nvars(dsid)
   for i = 1,col_count do
      local vname = sas.varname(dsid, i)
      col_names[i] = vname
   end

   -- now get the values into a table to return
   local _rows =  {}
   while sas.next(dsid) do
      local _row =  {}
      for col_num = 1,col_count do
         local vname = col_names[col_num]
         local value = sas.get_value(dsid, col_num)
         if trim_string_values and
            type(value) == "string" then
            value = stringutils.trim(value)
         end
         _row[vname] = value
      end
      _rows[#_rows+1] = _row
   end
   
   sas.close(dsid)
   return _rows
end

--- Queries a single row and returns a lua table 
-- @param selectStmt SQL select statement string with substition fields, should select one row
-- @param substitution_args Table of name value pairs to substitute in the sql_stmt string
-- @param[opt] trim_string_values Boolean to trim character values in the returned lua table
-- @param[opt] addl_lookup_level Additional stack levels to resolve variables for substitution. 0 is the default
--@return Table for the row. If no row is selected an error will be thrown
function M.select_one_row(selectStmt, substitution_args, trim_string_values, addl_lookup_level)
   addl_lookup_level=addl_lookup_level or 0
   local rows = M.select(selectStmt, substitution_args, trim_string_values, addl_lookup_level+1) -- add this function as a single level to what caller passed
   if #rows~=1 then
      errors.throw('utils_sasdata_other_than_one_row_returned', tostring(#rows))
   end
   return rows[1] -- return the inner row (set of named columns in a lua table)
end

--- Returns an array of column values from the query
-- @param selectStmt SQL select statement string with substition fields, should select one row
-- @param substitution_args Table of name value pairs to substitute in the sql_stmt string
-- @param[opt] trim_string_values Boolean to trim character values in the returned lua table
-- @param[opt] addl_lookup_level Additional stack levels to resolve variables for substitution. 0 is the default
--@return Table for the row. If no row is selected an error will be thrown
--@usage
--local match = sasdata.select_one_column([[select class_var_list_sk from @rrr_project_libref@.class_var_list_x_class_var
--   where class_var_list_no = @nclass@ and class_var_order_no = @order_no@ and class_var_sk = @sk@]], 
--   {rrr_project_libref=rrr_project_libref, nclass=#keys_inorder, order_no=#keys_inorder, sk=keys_inorder[#keys_inorder]},
--   false)
function M.select_one_column(selectStmt, substitution_args, trim_string_values, addl_lookup_level)
   addl_lookup_level=addl_lookup_level or 0
   trim_string_values = trim_string_values or true  -- default to true if not specified

   local dsname = "WORK._select_into_array"
   M.delete_ds(dsname)
   
   M.sql("create table " .. dsname .. " as\n" ..
        selectStmt .. ";",   -- add a benign semicolon in case the caller didn't include one
       substitution_args, nil, addl_lookup_level+1) -- add this function as a single level to what caller passed

   local dsid = M.open(dsname)

   local col_count = sas.nvars(dsid)
   assert(col_count == 1, "Query was expected to return 1 column but got " .. tostring(col_count) .. " columns")

   -- now get the values into a table to return
   local _column =  {}
   while sas.next(dsid) do
      local value = sas.get_value(dsid, 1)
      if trim_string_values and
         type(value) == "string" then
         value = stringutils.trim(value)
      end
      _column[#_column+1] = value
   end
   sas.close(dsid)
   M.delete_ds(dsname)
   return _column
end

--- Select a single value from a data set
-- @param selectStmt SQL select statement string with substition fields, should select one row
-- @param substitution_args Table of name value pairs to substitute in the sql_stmt string
-- @param[opt] trim_string_values Boolean to trim character values in the returned lua table
-- @param[opt] addl_lookup_level Additional stack levels to resolve variables for substitution. 0 is the default
--@return A string or a number. If no value is selected an error will be thrown
function M.select_one_value (selectStmt, substitution_args, trim_string_values, addl_lookup_level)
   addl_lookup_level=addl_lookup_level or 0
   local value = M.select_one_column(selectStmt, substitution_args, trim_string_values, addl_lookup_level+1) -- add this function as a single level to what caller passed
   if #value == 0 then
     return nil
   elseif #value > 1 then
      errors.throw('utils_sasdata_more_than_one_value_returned', tostring(#value))
   end
   return value[1]
end


--- Get the number of observation in a data set
--@param dsname Fully qualified name of the data set
--@return number of observations in the data set
function M.get_row_count(ds)
   local dsid = M.open(ds)
   local out_value = dsid:nobs() -- TODO: fix this: it should be nlobs but that function is not defined yet (nobs really does give back numbers including deleted records)
   dsid:close()
   return out_value
end

--- Set a value in a column of a data set
--@param dsname Fully qualified name of the data set
--@param colname Name of the column
--@param value The value to set
--@param whereclause A string containing the where clause to select the rows
--@return If PROC SQL diagnoses an error, a Lua error is thrown
function M.set_column_value(ds, col, value, filter)
   -- use sql instead of datastep since data step drops integrity constraints
   local value_syntax = value
   if tostring(value) == "string" then
      value_syntax = '"' .. value_syntax .. '"'
   end
   
   local where_clause = ""
   if filter and #filter > 0 then
      where_clause = 'where ' .. filter
   end
   
   M.sql[[update @ds@
         set @col@=@value_syntax@
         @where_clause@
      ]] 
end
    


local Rowset = {}
M.Rowset = Rowset

function Rowset:new(o)
   o = o or {}
   setmetatable(o, self)
   self.__index = self
   return o
end
function Rowset:close()
   if self.dsid then
      self.dsid:close()
      sas.delete(self.view,"view")
      self.dsid = nil
   end
end

function Rowset:rows()
   return coroutine.wrap(function()

                            self.view = "WORK.T" .. tostring(sas.datetime()):gsub("%.",""):sub(-4)
                            sas.submit([[
                              proc sql noprint;
                              create view @name@ as @query@;
                              quit;]], {name=self.view, query=self.query})
                            assert(sas.exists(self.view, "view"))
                            local dsid = M.open(self.view)
                            self.dsid  = dsid
                            for row in dsid:rows() do
                               coroutine.yield(row)
                            end
                            self:close()
                         end)

end

--- Run an SQL query and return an iterateable object
--- @usage
-- local rowset = sasdata.query("select age, height from sashelp.class")
-- for row in rowset:rows() do
--    print(row.age, row.height)
-- end
-- rowset:close()
function M.query(query)
   return Rowset:new({query=query})
end

--- Get the list of varables defined in data set
--@param dsname Fully qualified name of the data set
--@return an array of variable names
function M.varlist(dsname)
   if not string.find(dsname, "%.") then
      dsname = "WORK." .. dsname
   end
   local dsid = M.open(dsname,'i')
   local vars = {}
   local nvars = sas.nvars(dsid)
   for n = 1,nvars do
      local vname = sas.varname(dsid, n)
      table.insert(vars, vname)
   end
   dsid:close()
   return vars
end

--- Deletes all rows from a data set, keeping existing indexes, etc. build on the data set.
-- Note that other ways to delete rows can remove indexes as a side-effect, this way does not.
-- @param dsname Fully qualified name of the data set
function M.delete_all_rows(dataset)
   M.sql([[delete from @dataset@;]])
end


local recursively_delete_all_rows_callback=function(dir_id, type, memname, level, parent, context, arg)
   if type=='D' then
      M.member_list(libref)
   end
end

--- Deletes all rows from all data sets in a directory, keeping existing indexes, etc. build on the data set.
-- Note that other ways to delete rows can remove indexes as a side-effect, this way does not.
-- @param dir Fully qualified directory path
function M.recursively_delete_all_rows(dir)
  filesys.dir_tree_walk(dir, recursively_delete_all_rows_callback)
end


return M
