local playpen = require 'sas.risk.rmx.playpen'
local rrr = require 'sas.risk.rmx.rrr'
local reportregistry = require 'sas.risk.rmx.reportregistry'
local psdm = require 'sas.risk.rmx.psdm' 
local errors = require 'sas.risk.utils.errors'
local filesys = require 'sas.risk.utils.filesys'
local log4sas = require 'sas.risk.rmx.log4sas'
local sasdata = require 'sas.risk.utils.sasdata'
local lock_manager = require 'sas.risk.utils.lock_manager'

local M = {}

local User = {}
local UserEntityContext = {}
M.User = User
M.UserEntityContext = UserEntityContext


--- Strip away Windows domain qualifiers   
local function normalize(login_id)
   login_id = string.gsub(login_id, "^.*\\","")
   login_id = string.gsub(login_id, "@.*$","")
   return login_id
end

-- static functions
local use_entity = function(product, entity, username, interactive)
  if not entity or #entity < 1 then
       entity = "main"
   end
   
   -- strip away Windows domain qualifiers                 
   username = normalize(username)
   sas.setoption("compress","binary")
   sas.setoption("threads","1")
   uec = UserEntityContext:new{product=product, entity=entity, username=username, interactive=interactive}
   uec:init()
   sas.gsymput("_RSK_LAST_LOGGED_IN_USERNAME", username)
   sas.gsymput("_RSK_LAST_ENTITY", entity)
   return uec
end
M.use_entity = use_entity



function User:new(o)
   o = o or {}
   setmetatable(o, self)
   self.__index = self
   return o
end

function User:get_login_for_domain(domain)
   return self.identity.logins[domain]
end



function UserEntityContext:new(o)
   o = o or {}
   setmetatable(o, self)
   self.__index = self
   return o
end

local function is_usermatch(table, loginid) 
   if not table.attrs.UserID then
      return false
   end
   local normalized = normalize(table.attrs.UserID)
   return normalized == loginid
end

function UserEntityContext:load_user()
   local loginid = self.username
   if not loginid then
      loginid = normalize(sas.getoption("metauser"))
   end

   local identity = {}
   if (pcall(function() identity = self.product.xomi:get_identity() end)) then
   -- find the login in metadata that matches loginid (the user id), in order to determine the authentication domain
      local domainName
      local loginidUpper = loginid:upper()
      for domain,login in pairs(identity.logins) do
         local userid = normalize(login.userid):upper()
         if userid == loginidUpper then
            domainName = domain
            break
         end
      end
   else
      identity.logins = {}
      identity.name = loginid
      identity.display_name = loginid
      identity.groups = {}
   end

   local password = tostring(sas.symget("RSK_PASSWORD"))
   sas.gsymput("RSK_PERSON", identity.name)
   sas.gsymput("RSK_USERID", loginid)
   self.user = User:new { loginid=loginid,
                 authdomain=domainName,
                 foundation=foundation,
                 password=password,
                 identity=identity
              }

end


function UserEntityContext:ensure_playpen_area()
  filesys.mkdir(self:path_for("playpens"))
end


function UserEntityContext:ensure_rrr()
  local rr = rrr.Rrr:new{libref="RD_RPT", uec=self}
  rr:init()
  if not rr:exists() then
      rr:create()
  end
end

function UserEntityContext:ensure_report_registry()
  self.report_registry:ensure()
end

function UserEntityContext:ensure_cube_area()
  filesys.mkdir( self:path_for("cubes"))
end

function UserEntityContext:ensure_job_history()
   -- allocate _RMXJOBH
   local jh_path = self:path_for("job_history")
   rsk.libname("_RMXJOBH", jh_path, {ensure=1})

   rsk.ensure_table("_RMXJOBH.JOBS", {
          { name="id", type="C", length=44},
          { name="name", type="C", length=100 },
          { name="entity", type="C", length=32},
          { name="playpen", type="C", length=256},
          { name="application", type="C", length=32},
          { name="started", type="N", format="nldatm21." },
          { name="ended", type="N" ,format="nldatm21." },
          { name="status", type="C",length=21 },
          { name="configuration_version_id", type="C",length=16}
    })

    rsk.ensure_table("_RMXJOBH.TASKS", {
         {name="job_id", type="C", length=44},
         {name="id", type="C", length=44},
         {name="name", type="C", length=100 },
         {name="started", type="N", format="nldatm21." },
         {name="ended", type="N" ,format="nldatm21." },
         {name="status", type="C",length=21 },
         {name="task_cd", type="C",length=50}
    })



   rsk.ensure_table("_RMXJOBH.ARGS", {
      { name="job_id", type="C", length=44},
      { name="task_id", type="C", length=44},
      { name="name", type="C", length=100 },
      { name="value", type="C", length=256}
    })
end

function UserEntityContext:ensure_metadata_folder_for_entity()
   print("ensuring metadata folder for entity... product folder name = " .. self.product.product_folder_name)
   if self.user.identity.my_folder_id then
      local uri = sas.symget("rsk_entity_meta_uri") 

      if uri==nil or string.len(uri) < 1 then
         rsk.submit([[
                          %rsk_ensure_folder_for_entity(entity=@entity@, username=@username@, l_application=@product_folder_name@, 
                                                        user_home_folder=@home_folder_id@,
                                                        outvar=rsk_entity_meta_uri);
                    ]],{username=self.username, 
                        entity=self.name, product_folder_name=self.product.product_folder_name, 
                        home_folder_id=self.user.identity.my_folder_id})
      end
   else
      print("Could not ensure metadata folder - user home folder not found.")
   end
end

function UserEntityContext:ensure_lock_dir(libref, location)
   local dir = self:path_for(location)
   lock_manager.ensure_lock_directory(libref, dir)
end


function UserEntityContext:ensure_jobs_dir()
   local dir = self:path_for("jobs")
   filesys.mkdir(dir)
end

function UserEntityContext:ensure_ui_work_area()
   local dir = self:path_for("ui_work_area")
   rsk.libname("_RMXWORK",  dir, {ensure=1})
end

function UserEntityContext:ensure()
   -- for the sake of macros that still get called
   sas.print("%3zValidating user-entity context.")
   self:ensure_lock_dir("RD_SLOCK", "system_locks")
   self:ensure_lock_dir("RD_ULOCK", "user_locks")
   self:ensure_system_data()
   self:allocate_rrr()
   self:ensure_rrr()

   self:ensure_report_registry()

   self:ensure_playpen_area()
   self:ensure_cube_area()
   self:ensure_job_history()

   sas.print("%3zUser entity context is valid.")
end

function UserEntityContext:init()
   self:load_user()

   local root_dir = self.product:get_property("risk.entity.root")
   Debug('Entity root is: ' .. tostring(root_dir))

   self.root_dir = root_dir
   self.name = string.lower(self.entity)
   self.user_home = root_dir .. "/userdata/" .. self.username
   self.home = self.user_home .. "/entities/" .. self.name
   self.entity_home = root_dir .. "/datamarts/" .. self.name

   self.log_dir = self.user_home .. "/interactive_logs/"
   local sessionLog = sas.symget("_SESSIONLOG")

   filesys.mkdir(self.log_dir)
   local log_rollover = false
   if sessionLog and sessionLog ~= "" then
      log_rollover = true
   end

   sas.gsymput("USER_HOME", self.user_home)
   sas.gsymput("USER_ENTITY_HOME", self.home)
   sas.gsymput("ENTITY_HOME", self.entity_home)
   sas.gsymput("ENTITY_ROOT", self.root_dir)
   
   if log_rollover then
      self:session_log_rollover(40)
   end

   -- save value for potential later use by recover_uec()
   local last_user = tostring(sas.symget("_RSK_LAST_LOGGED_IN_USERNAME"))

  local reports_area = self:path_for("reports")
  self.report_registry = reportregistry.ReportRegistry:new{path=reports_area, uec=self}

   if not last_user or last_user=="" then
      self:ensure()
   end
   self:ensure_metadata_folder_for_entity()
   self.psdm = psdm.PSDM:new{path=self:path_for("psdm"), uec=self}
   self.psdm:init()
   self:ensure_ui_work_area() -- _RMXWORK
end


local playpen_list_cb = function(dir_id, type, memname, level, parent, context, arg)
      -- only process directories
      if type ~= 'D' then
        return
      end

      -- current directory
      local dir = context .. "/" .. memname

      -- rule out directories that don't contain the _meta dataset
      local rc = rsk.libname("_tmp", dir)
      if not sas.exists("_tmp._meta") then 
         rsk.libname("_tmp")
         return
      end

      -- use previously set libref as flag to indicate whether or not
      -- the Shared Entity Datamart exists
      local sedm_exists = (sas.libref("_tmp0")==0)

      -- obtain current playpen status (locked, not locked)
      local status = 0

      if sas.fileexists(dir .. "/.pp.lck") then
         status = 1
      end

      -- read the _meta dataset to find out what type of playpen this is
      local dsid = sas.open('_tmp._meta')
      if not dsid then
         rsk.libname('_tmp')
         return
      end

      rc  = dsid:next()

      local dm_lib

      local type = dsid:get_value("type")
      dsid:close()
      
      if "S" == type then
          if sedm_exists then
             dm_lib = "_tmp0"
          end
      else  
         local pedm_dir = dir.."/datamart"
         if filesys.dir_exists(pedm_dir) then
            rc = rsk.libname("_tmp2", pedm_dir)
            if sas.exists("_tmp2.rundate") then
                dm_lib = "_tmp2"
            end    
         end
      end    

      sas.submit_([[
          data work._meta;
            length status 8 name $256 applications $ 50 rundatetime 8 configuration_set_id $ 32 
                   configuration_datetime 8 edm_creation_datetime 8 last_enrich_datetime 8 ;
                 format rundatetime nldatm21.;
                 ]], {})

      if dm_lib == nil then
           sas.submit_([[ set _tmp._meta;]], {})
      else
           sas.submit_([[ merge _tmp._meta @dm_lib@.rundate @dm_lib@.mart_info; ]], 
                      { dm_lib=dm_lib})
      end 

      sas.submit([[
           name = "@memname@";
           status = @status@;
           drop applications;
           run;
            /* append to our output table */
           proc append base=@arg@ data=work._meta force;
           run;
           proc sql noprint; drop table work._meta;quit;

           run;]], { memname = memname, status=status, arg=arg})
           
     if "_tmp2" == dm_lib then 
          sasdata.libname("_tmp2")
     end  

     sasdata.libname("_tmp")
 end


local user_list_cb = function(dir_id, type, memname, level, parent, context, arg)
   -- only process directories
   if type ~= 'D' then
      return
   end
   -- Each directory is a user, so append it to the output dataset
   sas.submit([[
      data work.temp;
         if 0 then set @arg@;
         user_name = "@memname@";
      run;
      proc append base=@arg@ data=work.temp force; run;
      proc datasets lib=work nolist nodetails; delete temp; quit;
   ]])   
end


function UserEntityContext:session_log_rollover(rollover_amount)
   rollover_amount = rollover_amount or 40
   local logs_path = self:path_for("interactive_logs")
   local fref = "_ilogs"
   sas.filename(fref, logs_path)
   local did = sas.dopen(fref)
   if did == 0 then
      -- could not open, so release the fileref
      sas.filename(fref)     
      return
   end
   local dnum = sas.dnum(did)
   if dnum > rollover_amount then 
      local num_to_delete = dnum - rollover_amount
      local file_names = {}
      for i=1, dnum do
         table.insert(file_names, sas.dread(did, i))
      end
      table.sort(file_names)
      
      for i=1, num_to_delete do
         local full_path = logs_path .. "/" .. file_names[i]
         sas.filename("_T", full_path)
         sas.fdelete("_T")
         sas.filename("_T")
      end
   end
   sas.dclose(did)
   sas.filename(fref)
end


function UserEntityContext:get_playpens(out_ds) 
 local sedm_path = self:path_for("shared_datamart")

  -- check if Shared Entity Data Mart exists and leave _tmp0 allocated if it does
  if filesys.dir_exists(sedm_path) then
   sasdata.libname("_tmp0", sedm_path)
   if not sas.exists("_tmp0.rundate") then
      sasdata.libname("_tmp0")
   end
 end

 -- create the output table
 rsk.ensure_table(out_ds, {
    { name="name", type="C", length=256},
    { name="status", type="N"},
    { name="type", type="C", length=1},
    { name="created", type="N", format="nldatm21."},
    { name="accessed", type="N", format="nldatm21."},
    { name="description", type="C", length=256},
    { name="application", type="C" ,length=32},
    { name="rundatetime", type="N", format="nldatm21."},
    { name="configuration_set_id", type="C", length=32},
    { name="configuration_datetime", type="N", format="nldatm21."},
    { name="edm_creation_datetime", type="N", format="nldatm21."},
    { name="last_enrich_datetime", type="N", format="nldatm21."},
    { name="last_saved_config_version_id", type="C", length=16}})

 -- walk the playpens location for this user, invoking the playpen_list_cb function
 -- for each item found in that directory
 filesys.dir_tree_walk(UEC:path_for("playpens"), playpen_list_cb, 1, out_ds)

 -- deallocate SEDM library if it was left allocated 
   sasdata.libname("_tmp0")
end

function UserEntityContext:get_playpen(playpen_name)
   local playpen_area = self:path_for("playpens")
   local playpen_dir = playpen_area .. "/" .. playpen_name

   assert(sas.fileexists(playpen_dir),"rsk_playpen_does_not_exist")

   local playpen = playpen.Playpen:new{uec=self, name=playpen_name,
                               location=playpen_dir, product=self.product}
   return playpen
end

local function copy_values(in_ds, out_ds, vars)
   for i,var in ipairs(vars) do
      local value = sas.get_value(in_ds, var)
      sas.put_value(out_ds, var, sas.get_value(in_ds, var))
   end
end


--- allocates shared dm library if it exists
local function dm_info(dm_path)
   local dm_info = nil

   if sas.fileexists(dm_path) then
      sasdata.libname("_tmp0", dm_path)
      if sas.exists("_tmp0.rundate") then
         -- load in the information
         local rundate_dsid = sas.open("_tmp0.rundate")
         local rundate
         if rundate_dsid and  sas.next(rundate_dsid) then
            rundate = sas.get_value(rundate_dsid, "rundatetime")
            sas.close(rundate_dsid)

            local mart_info_dsid = sas.open("_tmp0.mart_info")
            if mart_info_dsid and sas.next(mart_info_dsid) then
                   dm_info = {}
                   dm_info.rundatetime = rundate

                   dm_info.configuration_set_id = sas.get_value(mart_info_dsid, "configuration_set_id")
                   dm_info.configuration_datetime = sas.get_value(mart_info_dsid,"configuration_datetime")
                   dm_info.edm_creation_datetime = sas.get_value(mart_info_dsid, "edm_creation_datetime")
                   dm_info.last_enrich_datetime = sas.get_value(mart_info_dsid, "last_enrich_datetime")
                   sas.close(mart_info_dsid)
             end
         end
      end
      sasdata.libname("_tmp0")
   end
   return dm_info
end

function UserEntityContext:playpens(out_ds)
   local out = out_ds or "work._out"
   local playpens_path = self:path_for("playpens")
   local playpens_logical = sasxx.assign(playpens_path)
   local playpens_dir = playpens_logical:opendir()
   rsk.ensure_table(out, {
                   {name="name", type="C", length=256},
                   {name="status", type="N"},
                   {name="description", type="C",length=256},
                   {name="application", type="C", length=32},
                   {name="type", type="C", length=1},
                   {name="created", type="N", format="NLDATM21."},
                   {name="accessed", type="N", format="NLDATM21."},
                   {name="rundatetime", type="N", format="NLDATM21."},
                   {name="configuration_set_id", type="C", length=32},
                   {name="configuration_datetime", type="N", format="NLDATM21."},
                   {name="edm_creation_datetime", type="N", format="NLDATM21."},
                   {name="last_enrich_datetime", type="N", format="NLDATM21."},
                   {name="last_saved_config_version_id", type="C", length=16}
             })
   local out_dsid = sasdata.open(out, "u")

   -- check if shared datamart exists. If it does, allocate a temporary library to it

   local sdm_values = dm_info(self:path_for("shared_datamart"))

   for dirent in playpens_dir:read() do
      local dm_values
      if (dirent.name ~= "." and dirent.name ~= "..") then
         local path  = playpens_path .. "/" .. dirent.name
         local status  = 0
         if sas.fileexists(path.."/.pp.lck") then
            status = 1
         end
         sasdata.libname("_tmp", path)
         if sas.exists("_tmp._meta") then
            local dsid = sas.open("_tmp._meta")
            if dsid > 0 then
               if sas.next(dsid) then
                  sas.append(out_dsid)
                  if sas.get_value(dsid, "type") == "P" then
                     dm_values = dm_info(path .. "/datamart")
                  else
                     dm_values = sdm_values
                  end

                  copy_values(dsid, out_dsid, {"type", "application", "created", "accessed", "description"})
                  if dm_values then
                     for k,v in pairs(dm_values) do
                        sas.put_value(out_dsid, k, v)
                     end
                  end

                  sas.put_value(out_dsid, "name", dirent.name)

                  sas.put_value(out_dsid,"status", status)
                  sas.update(out_dsid)
               else
                  print("Could not read playpen table - skipping ", dirent.name)
               end
               dsid = sas.close(dsid)
            end
         end
      end
   end
   sas.close(out_dsid)
   sasdata.libname("_tmp")

   playpens_dir:close()
   playpens_logical:deassign()

   if not out_ds then
      local result = sas.load_ds(out)
      sas.delete(out)
      return result
   end
end

function UserEntityContext:entity_home()
  return self.root_dir .. "/datamarts/" .. self.name
end

--- 
--  Call the provided function with its parameters inside a user-scoped critical code section named by the caller and 
--  enforced by a user-scoped write lock on that section name, returning the function's results. 
--  Any exception thrown while executing the function is re-thrown by this function.
--  
--  All SAS processes for this RSK user & critical_code_name combination will execute the given function serially
--  when called through here.
function UserEntityContext:protected_call_for_critical_code(critical_code_name, ftn, ...)
   return lock_manager.protected_call_with_lock('rd_ulock', critical_code_name, "W", ftn, ...)
end


--- Common function to provide a call to a given function, locking some resource, and unlocking it after
-- the call.  This call captures any exceptions that are thrown and frees the acquired lock prior to 
-- rethrowing the exception.  
function UserEntityContext:protected_call_with_user_lock(object_name, lock_type, ftn, ...)
   return lock_manager.protected_call_with_lock('rd_ulock', object_name, lock_type, ftn, ...)
end


function UserEntityContext:path_for(item)
    if "reportmart" == item then
       return self.user_home .. "/reportmart"
    elseif "system_data" == item then
       return self.root_dir ..  "/sysdata"
    elseif "reports" == item then
        return self.home .. "/reports"
    elseif "playpens" == item then
        return self.home .. "/playpens"
    elseif "cubes" == item then
        return self.home .. "/cubes"
    elseif "job_history" == item then
        return self.user_home .. "/job_history"
    elseif "shared_datamart" == item then
       return self.entity_home
    elseif "psdm" == item then
       return self.user_home .. "/runtime_configs"
    elseif "system_locks" == item then
       return self.root_dir .. "/locks"
    elseif "jobs" == item then 
       return self.root_dir .. "/jobs"
    elseif "user_locks" == item then
       return self.user_home .. "/locks"
--    elseif "hp_stress_portfolio_cube" == item then
--       return self.root_dir .. "/hp/stress_cubes"  
    elseif "ui_work_area" == item then
       return self.user_home .. "/scratch"
    elseif "interactive_logs" == item then
       return self.user_home .. "/interactive_logs"
--    elseif "promoted_configs" == item then
--       return self.root_dir .. "/config/promoted_configs"
--    elseif "config_sets" == item then
--       return self.root_dir .. "/config/config_sets"
    end
    assert(UEC.current_playpen, "Playpen not open")
    local playpen_root = UEC.current_playpen.location
    if "snapshot" == item then
       return playpen_root .. "/cfg_snap"
    elseif "snapshot_temp" == item then
       return playpen_root .. "/cfg_snap_temp"
    elseif "datamart" == item then
       return playpen_root .. "/datamart"
    else errors.throw(tostring(item) .. " type=(" .. type(item) ..  ")  is ***NOT KNOWN ***")
    end
end


function UserEntityContext:allocate_rrr()
   local path = self:path_for("reportmart")
   rsk.libname("RD_RPT",  path, {ensure=1})
end


function UserEntityContext:ensure_system_data()
   local dir = self:path_for("system_data")

   rsk.libname("rd_sysda", dir, { ensure=true, readOnly=true })

   if not sas.exists("rd_sysda.conf_editability") then
      rsk.libname("rd_sysda", dir)
      sas.submit ([[
      %let misc_system_data_dir=%rsk_get_source_dir(component=common)/system_data;
      %rsk_create_sample_ds_in_dir(in_sample_ds_dir=&misc_system_data_dir.);
      ]],  { PREFIX=string.lower(self.product.product)})
      rsk.libname("rd_sysda", dir, { readOnly=true})
   end
end

--- Macro entry points


---
-- rsk_open_playpen
--
local function open_playpen(playpen_name)
    local playpen = UEC:get_playpen(playpen_name)
    playpen:open()
    return playpen    
end
M.open_playpen = open_playpen


---
-- rsk_use_playpen
--

local function use_playpen(playpen_name, status_out)
   local pp = UEC:get_playpen(playpen_name)
   pp:use(status_out)
   return pp
end
M.use_playpen = use_playpen

---
-- rsk_close_playpen
--
local function close_playpen()
    local playpen_name = sas.symget("_RSK_LAST_PLAYPEN")
    assert(playpen_name, "No playpen open")
    local playpen = UEC:get_playpen(playpen_name)
    playpen:close()
end
M.close_playpen = close_playpen


local function get_playpens(out_ds)
  UEC:get_playpens(out_ds)
end
M.get_playpens = get_playpens

-- Create a list of users and output it in a dataset
local function get_users(out_ds)
   local root_dir = UEC.product:get_property("risk.entity.root")
   
   -- Create the output table
   rsk.ensure_table(out_ds, {
      { name="user_name", type="C", length=256}})

   -- Walk the playpens location for this user, invoking the playpen_list_cb function
   --    for each item found in that directory
   filesys.dir_tree_walk(root_dir .. "/userdata/", user_list_cb, 1, out_ds)
end
M.get_users = get_users

-- Delete all playpens
-- Not to be used by user, only meant for developers
local function delete_all_playpens()
   local root_dir = UEC.product:get_property("risk.entity.root")

   -- Get list of users
   local users = "work._users"
   sasdata.delete_ds(users)
   get_users(users)
   local user_list = sasdata.select_one_column("select user_name from work._users")
   sasdata.delete_ds(users)

   -- Iterate through all users
   for i,user in ipairs(user_list) do

      -- Get list of entities for this user
      local entities_path = root_dir .. "/userdata/" .. user .. "/entities/"
      local fref = "entities"
      sas.filename(fref, entities_path)
      local did = sas.dopen(fref)
      if did ~= 0 then
         local dnum = sas.dnum(did)

         --Iterate through all entities for this user
         for j=1,dnum do
            local entity = sas.dread(did, j)
            if entity ~= "" and filesys.dir_exists(entities_path .. entity) then

               -- Initialize user-entity context
               _G.UEC = use_entity(UEC.product, entity, user, "N")

               -- Get a list of playpens for this user-entity pair
               local playpens = "work._playpens"
               sasdata.delete_ds(playpens)
               get_playpens(playpens)
               local playpen_list = sasdata.select_one_column("select name from work._playpens")
               sasdata.delete_ds(playpens)

               -- Iterate through the playpens, deleting them
               for k,curr_playpen in ipairs(playpen_list) do
                  sas.submit([[%rsk_delete_playpen(name=@curr_playpen@)]])
               end

            end -- entity check
         end -- entity loop

      end -- directory of entities check
      sas.dclose(did)
      sas.filename(fref)

   end -- user loop

end -- function delete_all_playpens
M.delete_all_playpens = delete_all_playpens

return M
